package com.winhc.data.push.service.impl;

import com.winhc.data.push.bean.*;
import com.winhc.data.push.common.Constant;
import com.winhc.data.push.configuration.ProjectParamInit;
import com.winhc.data.push.service.DataWorksService;
import com.winhc.data.push.service.DingTalkService;
import com.winhc.data.push.service.OdpsService;
import com.winhc.data.push.service.TouchService;
import com.winhc.data.push.utils.BaseUtils;
import com.winhc.data.push.utils.DateUtils;
import lombok.AllArgsConstructor;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;

import java.time.Duration;
import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;

import static com.winhc.data.push.common.Constant.TC_OSS_PRE;
import static com.winhc.data.push.utils.BaseUtils.getPrcessDate;
import static com.winhc.data.push.utils.DateUtils.getPrcessDateV2;

/**
 * @author π
 * @Description:
 * @date 2022/7/21 18:18
 */
@Service
@Slf4j
@AllArgsConstructor
public class TouchServiceImpl implements TouchService {

    private final DataWorksService dataWorksService;
    private final DingTalkService dingTalkService;
    private final Map<String, DataWorksFlowTask> dataWorksFlowTaskMap;

    private final OdpsService odpsService;

    @Override
    public Boolean export2Mongo(String project, String ds) {
        return query(project, dataWorksService.export2Mongo(ds).getReturnValue());
    }

    @Override
    public Boolean export2OSS(String project, String ds) {
        return query(project, dataWorksService.export2OSS(ds).getReturnValue());
    }

    /**
     * 58数据导出
     *
     * @param project
     * @param ds
     * @return
     */
    @Override
    public Boolean export2OSSTc(String project, String ds) {
        return this.export2OSSTc(project, ds, null);
    }

    @Override
    public Boolean export2OSSTc(String project, String ds, List<String> tns) {
        String prcessDate = getPrcessDateV2(ds, 7);
        //当日分区数据
        List<String> ds_tns = odpsService.query(Constant.show_tn_sql, ds);
        //需求输出数据
        List<String> out_tns = ds_tns.stream()
                .filter(tn -> Constant.tc_tns.contains(tn))
                .distinct()
                .collect(Collectors.toList());
        //取交集
        if (tns != null && tns.size() > 0) {
            out_tns.retainAll(tns);
        }
        Map<Long, String> task_ids = out_tns.stream().map(tn -> {
            Map<String, String> otherParams = new HashMap<>();
            otherParams.put("tn", tn);
            otherParams.put("ossPrefix", TC_OSS_PRE);
            otherParams.put("prcessDate", prcessDate);
            otherParams.put("detail", BaseUtils.getDetailName(ds, tn));
            Long returnValue = dataWorksService.export2OSS(ds, ProjectParamInit.exportOssFlowNameTC,
                    DateUtils.parseDate(ds),
                    dataWorksFlowTaskMap.get(ProjectParamInit.exportOssBeanNameTC),
                    otherParams).getReturnValue();
            return Arrays.asList(returnValue, tn);
        }).distinct().collect(Collectors.toMap(c -> Long.parseLong(c.get(0).toString()), c -> c.get(1).toString(), (c, v) -> v));

        return queryList(project, task_ids);
    }

    @SneakyThrows
    @Override
    public Boolean queryList(String projectName, Map<Long, String> task_ids_maps) {
        LocalDateTime start = LocalDateTime.now();
        int totalTask = task_ids_maps.size();
        int i = 0;
        int successTask = 0;
        int failedTask = 0;
        ArrayList<Long> hours = new ArrayList<>();
        while (true) {
            int awaitTask = 0;
            boolean flag;

            for (Long id : task_ids_maps.keySet()) {
                Map<String, TaskFlowEnum> query = dataWorksService.query(projectName, id);
                Map<String, List<String>> status = query.entrySet().stream()
                        .map(e -> {
                            Entry<String, String> entry = new Entry<>();
                            if (e.getValue().equals(TaskFlowEnum.SUCCESS) || e.getValue().equals(TaskFlowEnum.FAILURE)) {
                                entry.setKey(e.getValue().getMsg());
                            } else {
                                entry.setKey("运行中");
                            }
                            entry.setValue(e.getKey());
                            return entry;
                        }).collect(Collectors.groupingBy(Entry::getKey, Collectors.mapping(Entry::getValue, Collectors.toList())));
                List<String> failure = status.getOrDefault(TaskFlowEnum.FAILURE.getMsg(), new ArrayList<>());
                List<String> await = status.getOrDefault(TaskFlowEnum.RUNNING.getMsg(), new ArrayList<>());

                if (failure.size() != 0) {
                    failedTask++;
                    log.error("导出任务失败 tn : {} , id : {}", task_ids_maps.getOrDefault(id, null), id);
                    dingTalkService.sendMessage("step1 导出数据报错, \n tn: " + task_ids_maps.getOrDefault(id, null) + " !!!!!!");
                } else {
                    if (await.size() != 0) {
                        awaitTask++;
                        LocalDateTime now = LocalDateTime.now();
                        Duration duration = Duration.between(start, now);
                        long hour = duration.toHours();
                        if (hour >= 1 && !hours.contains(hour)) {
                            i++;
                            hours.add(hour);
                            log.info("导出数据阻塞告警 !!!!!!");
                            dingTalkService.sendMessage("导出数据阻塞告警, \n tn: " + task_ids_maps.getOrDefault(id, null) + " !!!!!!");
                        }
                    } else {
                        successTask++;
                    }
                }
            }

            flag = awaitTask == 0 || failedTask != 0;
            //超时退出
            if (DateUtils.getCurrentHours() >= 22 || i > 6) {
                dingTalkService.sendMessage("导出数据超时, 退出 !!!!!!");
                log.info("超时程序主动退出 !!!!!!");
                return false;
            }
            log.info("\nawait task:{}\ntotal task:{}\nsuccess task:{}\nfailure task:{}\n"
                    , awaitTask
                    , totalTask
                    , totalTask - awaitTask - failedTask
                    , failedTask
            );
            if (flag) {
                break;
            }
            Thread.sleep(20 * 000);
        }
        return true;
    }

    @SneakyThrows
    @Override
    public Boolean query(String projectName, Long dagId) {
        LocalDateTime start = LocalDateTime.now();
        int totalTask = 1;
        int i = 0;
        int successTask = 0;
        int failedTask = 0;
        ArrayList<Long> hours = new ArrayList<>();
        while (true) {
            int awaitTask = 0;
            List<String> empty = new ArrayList<>();
            Map<String, TaskFlowEnum> query = dataWorksService.query(projectName, dagId);
            Map<String, List<String>> status = query.entrySet().stream()
                    .map(e -> {
                        Entry<String, String> entry = new Entry<>();
                        if (e.getValue().equals(TaskFlowEnum.SUCCESS) || e.getValue().equals(TaskFlowEnum.FAILURE)) {
                            entry.setKey(e.getValue().getMsg());
                        } else {
                            entry.setKey("运行中");
                        }
                        entry.setValue(e.getKey());
                        return entry;
                    }).collect(Collectors.groupingBy(Entry::getKey, Collectors.mapping(Entry::getValue, Collectors.toList())));
            List<String> failure = status.getOrDefault(TaskFlowEnum.FAILURE.getMsg(), empty);
            List<String> await = status.getOrDefault(TaskFlowEnum.RUNNING.getMsg(), empty);

            if (failure.size() != 0) {
                log.error("任务失败 node : {}", dagId);
                return false;
            } else {
                if (await.size() != 0) {
                    awaitTask++;
                    LocalDateTime now = LocalDateTime.now();
                    Duration duration = Duration.between(start, now);
                    long hour = duration.toHours();
                    if (hour >= 1 && !hours.contains(hour)) {
                        i++;
                        hours.add(hour);
                        log.info("导出数据阻塞告警 !!!!!!");
                        dingTalkService.sendMessage("导出数据阻塞告警 !!!!!!");
                    }
                } else {
                    return true;
                }
            }

            //超时退出
            if (DateUtils.getCurrentHours() >= 22 || i > 6) {
                log.info("超时程序主动退出 !!!!!!");
                return false;
            }
            log.info("\nawait task:{}\ntotal task:{}\nsuccess task:{}\nfailure task:{}\n"
                    , awaitTask
                    , totalTask
                    , successTask
                    , failedTask
            );
            Thread.sleep(10000);
        }
    }

}
